==> v10.5.1.zip: brewman/overlord/src/app/journal/journal.component.ts <==
import { AfterViewInit, Component, ElementRef, OnInit, OnDestroy, ViewChild } from '@angular/core';
import { FormControl, FormGroup } from '@angular/forms';
import { MatAutocompleteSelectedEvent } from '@angular/material/autocomplete';
import { MatDialog } from '@angular/material/dialog';
import { ActivatedRoute, Router } from '@angular/router';
import { Hotkey, HotkeysService } from 'angular2-hotkeys';
import { round } from 'mathjs';
import * as moment from 'moment';
import { BehaviorSubject, Observable, of as observableOf } from 'rxjs';
import { debounceTime, distinctUntilChanged, map, switchMap } from 'rxjs/operators';

import { AuthService } from '../auth/auth.service';
import { Account } from '../core/account';
import { AccountBalance } from '../core/account-balance';
import { AccountService } from '../core/account.service';
import { DbFile } from '../core/db-file';
import { Journal } from '../core/journal';
import { ToasterService } from '../core/toaster.service';
import { User } from '../core/user';
import { Voucher } from '../core/voucher';
import { VoucherService } from '../core/voucher.service';
import { ConfirmDialogComponent } from '../shared/confirm-dialog/confirm-dialog.component';
import { ImageDialogComponent } from '../shared/image-dialog/image-dialog.component';
import { ImageService } from '../shared/image.service';
import { MathService } from '../shared/math.service';

import { JournalDataSource } from './journal-datasource';
import { JournalDialogComponent } from './journal-dialog.component';

@Component({
  selector: 'app-journal',
  templateUrl: './journal.component.html',
  styleUrls: ['./journal.component.css'],
})
export class JournalComponent implements OnInit, AfterViewInit, OnDestroy {
  @ViewChild('accountElement', { static: true }) accountElement?: ElementRef;
  @ViewChild('dateElement', { static: true }) dateElement?: ElementRef;
  public journalObservable = new BehaviorSubject<Journal[]>([]);
  dataSource: JournalDataSource = new JournalDataSource(this.journalObservable);
  form: FormGroup<{
    date: FormControl<Date>;
    addRow: FormGroup<{
      debit: FormControl<number>;
      account: FormControl<string | null>;
      amount: FormControl<string>;
    }>;
    narration: FormControl<string>;
  }>;

  voucher: Voucher = new Voucher();
  account: Account | null;
  accBal: AccountBalance | null = null;

  displayedColumns = ['debit', 'account', 'amount', 'action'];

  accounts: Observable<Account[]>;

  constructor(
    private route: ActivatedRoute,
    private router: Router,
    private dialog: MatDialog,
    private hotkeys: HotkeysService,
    private toaster: ToasterService,
    public auth: AuthService,
    private math: MathService,
    public image: ImageService,
    private ser: VoucherService,
    private accountSer: AccountService,
  ) {
    this.account = null;
    this.form = new FormGroup({
      date: new FormControl(new Date(), { nonNullable: true }),
      addRow: new FormGroup({
        debit: new FormControl(1, { nonNullable: true }),
        account: new FormControl(''),
        amount: new FormControl('', { nonNullable: true }),
      }),
      narration: new FormControl('', { nonNullable: true }),
    });
    this.accBal = null;
    // Setup Account Autocomplete
    this.accounts = this.form.controls.addRow.controls.account.valueChanges.pipe(
      map((x) => (x !== null && x.length >= 1 ? x : null)),
      debounceTime(150),
      distinctUntilChanged(),
      switchMap((x) => (x === null ? observableOf([]) : this.accountSer.autocomplete(x))),
    );
  }

  ngOnInit() {
    this.route.data.subscribe((value) => {
      const data = value as { voucher: Voucher };

      this.loadVoucher(data.voucher);
    });
    this.hotkeys.add(
      new Hotkey(
        'f2',
        (): boolean => {
          setTimeout(() => {
            if (this.dateElement) {
              this.dateElement.nativeElement.focus();
            }
          }, 0);
          return false; // Prevent bubbling
        },
        ['INPUT', 'SELECT', 'TEXTAREA'],
      ),
    );
    this.hotkeys.add(
      new Hotkey(
        'ctrl+s',
        (): boolean => {
          if (this.canSave()) {
            this.save();
          }
          return false; // Prevent bubbling
        },
        ['INPUT', 'SELECT', 'TEXTAREA'],
      ),
    );
    this.hotkeys.add(
      new Hotkey(
        'ctrl+p',
        (): boolean => {
          if (this.voucher.id && !this.voucher.posted && this.auth.allowed('post-vouchers')) {
            this.post();
          }
          return false; // Prevent bubbling
        },
        ['INPUT', 'SELECT', 'TEXTAREA'],
      ),
    );
  }

  ngAfterViewInit() {
    this.focusAccount();
  }

  ngOnDestroy() {
    this.hotkeys.reset();
  }

  loadVoucher(voucher: Voucher) {
    this.voucher = voucher;
    this.form.setValue({
      date: moment(this.voucher.date, 'DD-MMM-YYYY').toDate(),
      addRow: {
        debit: 1,
        account: null,
        amount: '',
      },
      narration: this.voucher.narration,
    });
    this.dataSource = new JournalDataSource(this.journalObservable);
    this.journalObservable.next(this.voucher.journals);
  }

  focusAccount() {
    setTimeout(() => {
      if (this.accountElement) {
        this.accountElement.nativeElement.focus();
      }
    }, 0);
  }

  addRow() {
    const formValue = this.form.value.addRow!;
    const amount = this.math.journalAmount(formValue.amount!, formValue.debit!);
    if (this.account === null || amount.amount === 0) {
      return;
    }
    const oldFiltered = this.voucher.journals.filter(
      (x) => x.account.id === (this.account as Account).id,
    );
    const old = oldFiltered.length ? oldFiltered[0] : null;
    if (old) {
      const a = old.debit * old.amount + amount.debit * amount.amount;
      old.debit = a < 0 ? -1 : 1;
      old.amount = Math.abs(a);
    } else {
      this.voucher.journals.push(
        new Journal({
          debit: amount.debit,
          amount: amount.amount,
          account: this.account,
          costCentre: null,
        }),
      );
    }
    this.journalObservable.next(this.voucher.journals);
    this.resetAddRow();
  }

  resetAddRow() {
    const amount = round(
      Math.abs(this.voucher.journals.map((x) => x.debit * x.amount).reduce((p, c) => p + c)),
      2,
    );
    this.form.controls.addRow.reset({
      account: null,
      amount: `${amount}`,
    });
    this.account = null;
    this.accBal = null;
    setTimeout(() => {
      if (this.accountElement) {
        this.accountElement.nativeElement.focus();
      }
    }, 0);
  }

  editRow(row: Journal) {
    const dialogRef = this.dialog.open(JournalDialogComponent, {
      width: '750px',
      data: {
        journal: { ...row },
        date: moment(this.form.value.date).format('DD-MMM-YYYY'),
      },
    });

    dialogRef.afterClosed().subscribe((result: boolean | Journal) => {
      if (!result) {
        return;
      }
      const j = result as Journal;
      if (
        j.account.id !== row.account.id &&
        this.voucher.journals.filter((x) => x.account.id === j.account.id).length
      ) {
        return;
      }
      Object.assign(row, j);
      this.journalObservable.next(this.voucher.journals);
      this.resetAddRow();
    });
  }

  deleteRow(row: Journal) {
    this.voucher.journals.splice(this.voucher.journals.indexOf(row), 1);
    this.journalObservable.next(this.voucher.journals);
    const amount = Math.abs(
      this.voucher.journals.map((x) => x.debit * x.amount).reduce((p, c) => p + c, 0),
    );
    this.form.controls.addRow.patchValue({
      amount: `${amount}`,
    });
  }

  canSave() {
    if (!this.voucher.id) {
      return true;
    }
    if (this.voucher.posted && this.auth.allowed('edit-posted-vouchers')) {
      return true;
    }
    return (
      this.voucher.user.id === (this.auth.user as User).id ||
      this.auth.allowed("edit-other-user's-vouchers")
    );
  }

  post() {
    this.ser.post(this.voucher.id as string).subscribe(
      (result) => {
        this.loadVoucher(result);
        this.toaster.show('Success', 'Voucher Posted');
      },
      (error) => {
        this.toaster.show('Danger', error);
      },
    );
  }

  save() {
    const voucher: Voucher = this.getVoucher();
    this.ser.saveOrUpdate(voucher).subscribe(
      (result) => {
        this.toaster.show('Success', '');
        if (voucher.id === result.id) {
          this.loadVoucher(result);
        } else {
          this.router.navigate(['/journal', result.id]);
        }
      },
      (error) => {
        this.toaster.show('Danger', error);
      },
    );
  }

  getVoucher(): Voucher {
    const formModel = this.form.value;
    this.voucher.date = moment(formModel.date).format('DD-MMM-YYYY');
    this.voucher.narration = formModel.narration!;
    return this.voucher;
  }

  delete() {
    this.ser.delete(this.voucher.id as string).subscribe(
      () => {
        this.toaster.show('Success', '');
        this.router.navigate(['/journal'], { replaceUrl: true });
      },
      (error) => {
        this.toaster.show('Danger', error);
      },
    );
  }

  confirmDelete(): void {
    const dialogRef = this.dialog.open(ConfirmDialogComponent, {
      width: '250px',
      data: { title: 'Delete Voucher?', content: 'Are you sure? This cannot be undone.' },
    });

    dialogRef.afterClosed().subscribe((result: boolean) => {
      if (result) {
        this.delete();
      }
    });
  }

  displayFn(account?: Account): string {
    return account ? account.name : '';
  }

  accountSelected(event: MatAutocompleteSelectedEvent): void {
    const account = event.option.value;
    this.account = account;
    const date = moment(this.form.value.date).format('DD-MMM-YYYY');
    this.accountSer.balance(account.id as string, date).subscribe((v) => {
      this.accBal = v;
    });
  }

  zoomImage(file: DbFile) {
    this.dialog.open(ImageDialogComponent, {
      width: '750px',
      data: file.resized,
    });
  }

  deleteImage(file: DbFile) {
    const index = this.voucher.files.indexOf(file);
    this.voucher.files.splice(index, 1);
  }
}

==> v11.0.1.zip: brewman/overlord/src/app/journal/journal.component.ts <==
import { AfterViewInit, Component, ElementRef, OnInit, OnDestroy, ViewChild } from '@angular/core';
import { FormControl, FormGroup } from '@angular/forms';
import { MatAutocompleteSelectedEvent } from '@angular/material/autocomplete';
import { MatDialog } from '@angular/material/dialog';
import { ActivatedRoute, Router } from '@angular/router';
import { Hotkey, HotkeysService } from 'angular2-hotkeys';
import { round } from 'mathjs';
import * as moment from 'moment';
import { BehaviorSubject, Observable, of as observableOf } from 'rxjs';
import { debounceTime, distinctUntilChanged, switchMap } from 'rxjs/operators';

import { AuthService } from '../auth/auth.service';
import { Account } from '../core/account';
import { AccountBalance } from '../core/account-balance';
import { AccountService } from '../core/account.service';
import { DbFile } from '../core/db-file';
import { Journal } from '../core/journal';
import { ToasterService } from '../core/toaster.service';
import { User } from '../core/user';
import { Voucher } from '../core/voucher';
import { VoucherService } from '../core/voucher.service';
import { ConfirmDialogComponent } from '../shared/confirm-dialog/confirm-dialog.component';
import { ImageDialogComponent } from '../shared/image-dialog/image-dialog.component';
import { ImageService } from '../shared/image.service';
import { MathService } from '../shared/math.service';

import { JournalDataSource } from './journal-datasource';
import { JournalDialogComponent } from './journal-dialog.component';

@Component({
  selector: 'app-journal',
  templateUrl: './journal.component.html',
  styleUrls: ['./journal.component.css'],
})
export class JournalComponent implements OnInit, AfterViewInit, OnDestroy {
  @ViewChild('accountElement', { static: true }) accountElement?: ElementRef;
  @ViewChild('dateElement', { static: true }) dateElement?: ElementRef;
  public journalObservable = new BehaviorSubject<Journal[]>([]);
  dataSource: JournalDataSource = new JournalDataSource(this.journalObservable);
  form: FormGroup<{
    date: FormControl<Date>;
    addRow: FormGroup<{
      debit: FormControl<number>;
      account: FormControl<string | null>;
      amount: FormControl<string>;
    }>;
    narration: FormControl<string>;
  }>;

  voucher: Voucher = new Voucher();
  account: Account | null;
  accBal: AccountBalance | null = null;

  displayedColumns = ['debit', 'account', 'amount', 'action'];

  accounts: Observable<Account[]>;

  constructor(
    private route: ActivatedRoute,
    private router: Router,
    private dialog: MatDialog,
    private hotkeys: HotkeysService,
    private toaster: ToasterService,
    public auth: AuthService,
    private math: MathService,
    public image: ImageService,
    private ser: VoucherService,
    private accountSer: AccountService,
  ) {
    this.account = null;
    this.form = new FormGroup({
      date: new FormControl(new Date(), { nonNullable: true }),
      addRow: new FormGroup({
        debit: new FormControl(1, { nonNullable: true }),
        account: new FormControl(''),
        amount: new FormControl('', { nonNullable: true }),
      }),
      narration: new FormControl('', { nonNullable: true }),
    });
    this.accBal = null;
    // Setup Account Autocomplete
    this.accounts = this.form.controls.addRow.controls.account.valueChanges.pipe(
      debounceTime(150),
      distinctUntilChanged(),
      switchMap((x) => (x === null ? observableOf([]) : this.accountSer.autocomplete(x))),
    );
  }

  ngOnInit() {
    this.route.data.subscribe((value) => {
      const data = value as { voucher: Voucher };

      this.loadVoucher(data.voucher);
    });
    this.hotkeys.add(
      new Hotkey(
        'f2',
        (): boolean => {
          setTimeout(() => {
            if (this.dateElement) {
              this.dateElement.nativeElement.focus();
            }
          }, 0);
          return false; // Prevent bubbling
        },
        ['INPUT', 'SELECT', 'TEXTAREA'],
      ),
    );
    this.hotkeys.add(
      new Hotkey(
        'ctrl+s',
        (): boolean => {
          if (this.canSave()) {
            this.save();
          }
          return false; // Prevent bubbling
        },
        ['INPUT', 'SELECT', 'TEXTAREA'],
      ),
    );
    this.hotkeys.add(
      new Hotkey(
        'ctrl+p',
        (): boolean => {
          if (this.voucher.id && !this.voucher.posted && this.auth.allowed('post-vouchers')) {
            this.post();
          }
          return false; // Prevent bubbling
        },
        ['INPUT', 'SELECT', 'TEXTAREA'],
      ),
    );
  }

  ngAfterViewInit() {
    this.focusAccount();
  }

  ngOnDestroy() {
    this.hotkeys.reset();
  }

  loadVoucher(voucher: Voucher) {
    this.voucher = voucher;
    this.form.setValue({
      date: moment(this.voucher.date, 'DD-MMM-YYYY').toDate(),
      addRow: {
        debit: 1,
        account: null,
        amount: '',
      },
      narration: this.voucher.narration,
    });
    this.dataSource = new JournalDataSource(this.journalObservable);
    this.journalObservable.next(this.voucher.journals);
  }

  focusAccount() {
    setTimeout(() => {
      if (this.accountElement) {
        this.accountElement.nativeElement.focus();
      }
    }, 0);
  }

  addRow() {
    const formValue = this.form.value.addRow;
    if (formValue === undefined) {
      return;
    }
    const amount = this.math.journalAmount(formValue.amount, formValue.debit ?? 1);
    if (this.account === null || amount.amount === 0) {
      return;
    }
    const oldFiltered = this.voucher.journals.filter(
      (x) => x.account.id === (this.account as Account).id,
    );
    const old = oldFiltered.length ? oldFiltered[0] : null;
    if (old) {
      const a = old.debit * old.amount + amount.debit * amount.amount;
      old.debit = a < 0 ? -1 : 1;
      old.amount = Math.abs(a);
    } else {
      this.voucher.journals.push(
        new Journal({
          debit: amount.debit,
          amount: amount.amount,
          account: this.account,
          costCentre: null,
        }),
      );
    }
    this.journalObservable.next(this.voucher.journals);
    this.resetAddRow();
  }

  resetAddRow() {
    const amount = round(
      Math.abs(this.voucher.journals.map((x) => x.debit * x.amount).reduce((p, c) => p + c)),
      2,
    );
    this.form.controls.addRow.reset({
      debit: this.form.value.addRow?.debit ?? 1,
      account: null,
      amount: `${amount}`,
    });
    this.account = null;
    this.accBal = null;
    setTimeout(() => {
      if (this.accountElement) {
        this.accountElement.nativeElement.focus();
      }
    }, 0);
  }

  editRow(row: Journal) {
    const dialogRef = this.dialog.open(JournalDialogComponent, {
      width: '750px',
      data: {
        journal: { ...row },
        date: moment(this.form.value.date).format('DD-MMM-YYYY'),
      },
    });

    dialogRef.afterClosed().subscribe((result: boolean | Journal) => {
      if (!result) {
        return;
      }
      const j = result as Journal;
      if (
        j.account.id !== row.account.id &&
        this.voucher.journals.filter((x) => x.account.id === j.account.id).length
      ) {
        return;
      }
      Object.assign(row, j);
      this.journalObservable.next(this.voucher.journals);
      this.resetAddRow();
    });
  }

  deleteRow(row: Journal) {
    this.voucher.journals.splice(this.voucher.journals.indexOf(row), 1);
    this.journalObservable.next(this.voucher.journals);
    const amount = Math.abs(
      this.voucher.journals.map((x) => x.debit * x.amount).reduce((p, c) => p + c, 0),
    );
    this.form.controls.addRow.patchValue({
      amount: `${amount}`,
    });
  }

  canSave() {
    if (!this.voucher.id) {
      return true;
    }
    if (this.voucher.posted && this.auth.allowed('edit-posted-vouchers')) {
      return true;
    }
    return (
      this.voucher.user.id === (this.auth.user as User).id ||
      this.auth.allowed("edit-other-user's-vouchers")
    );
  }

  post() {
    this.ser.post(this.voucher.id as string).subscribe(
      (result) => {
        this.loadVoucher(result);
        this.toaster.show('Success', 'Voucher Posted');
      },
      (error) => {
        this.toaster.show('Danger', error);
      },
    );
  }

  save() {
    const voucher: Voucher = this.getVoucher();
    this.ser.saveOrUpdate(voucher).subscribe(
      (result) => {
        this.toaster.show('Success', '');
        if (voucher.id === result.id) {
          this.loadVoucher(result);
        } else {
          this.router.navigate(['/journal', result.id]);
        }
      },
      (error) => {
        this.toaster.show('Danger', error);
      },
    );
  }

  getVoucher(): Voucher {
    const formModel = this.form.value;
    this.voucher.date = moment(formModel.date).format('DD-MMM-YYYY');
    this.voucher.narration = formModel.narration ?? '';
    return this.voucher;
  }

  delete() {
    this.ser.delete(this.voucher.id as string).subscribe(
      () => {
        this.toaster.show('Success', '');
        this.router.navigate(['/journal'], { replaceUrl: true });
      },
      (error) => {
        this.toaster.show('Danger', error);
      },
    );
  }

  confirmDelete(): void {
    const dialogRef = this.dialog.open(ConfirmDialogComponent, {
      width: '250px',
      data: { title: 'Delete Voucher?', content: 'Are you sure? This cannot be undone.' },
    });

    dialogRef.afterClosed().subscribe((result: boolean) => {
      if (result) {
        this.delete();
      }
    });
  }

  displayFn(account?: Account | string): string {
    return !account ? '' : typeof account === 'string' ? account : account.name;
  }

  accountSelected(event: MatAutocompleteSelectedEvent): void {
    const account = event.option.value as Account;
    this.account = account;
    const date = moment(this.form.value.date).format('DD-MMM-YYYY');
    this.accountSer.balance(account.id as string, date).subscribe((v) => {
      this.accBal = v;
    });
  }

  zoomImage(file: DbFile) {
    this.dialog.open(ImageDialogComponent, {
      width: '750px',
      data: file.resized,
    });
  }

  deleteImage(file: DbFile) {
    const index = this.voucher.files.indexOf(file);
    this.voucher.files.splice(index, 1);
  }
}
